from io import IOBase
import os
from urllib.parse import urlparse
import base64
from .representations import Representation
from .target import Target


supported_formats = {
    'mol': ["cif", "cifcore", "pdb", "pdbqt", "gro", "xyz", "mol", "sdf", "mol2", "lammps_data", "lammps_traj_data"],
    'snapshot': ["json", "molj", "molx", "zip"],
    'coords': ["dcd", "xtc", "trr", "nctraj", "lammpstrj"],
}

def parse_molecule(inp, fmt=None, component=None, preset={'kind': 'standard'}):
    """
    Parse the molecule for `data` parameter of molstar viewer.

    Parameters
    ----------
    `inp` — str | file-like object
        The file path to molecule or the file content of molecule.

        If `inp` was set to file path, the format can be automatically determined.
        Otherwise the format has to be specified manually.
    `fmt` — str (optional)
        Format of the input molecule. 
        Supported formats include `cif`, `cifcore`, `pdb`, `pdbqt`, `gro`, `xyz`, `mol`, `sdf`, `mol2`, `lammps_data`, `lammps_traj_data` (default: `None`)
    `component` — dict | List[dict] (optional)
        Component to be created in molstar. 
        If not specified, molstar will use its default settings. (default: `None`)

        Use helper function `create_component` to generate correct data for this parameter.
    `preset` — dict (optional)
        The preset for molstar of how to display the loaded structure file.

    Returns
    -------
    `dict`
        The value for the `data` parameter.

    Raises
    ------
    `RuntimeError`
        If the input format is not supported by molstar viewer, raises RuntimeError.
    """
    # provided a filename as input
    if os.path.isfile(inp):
        # if format is not specified, infer from filename
        if not fmt:
            name, fmt = os.path.splitext(inp)
        with open(inp, 'r') as f:
            data = f.read()
    else:
        # provided a file-like object as input
        if isinstance(inp, IOBase):
            data = inp.read()
        # provided file content as input
        else:
            data = inp
        if type(data) == bytes: data = data.decode()
    if not fmt: raise RuntimeError("The format must be specified if you didn't provide a file name.")
    fmt = fmt.strip('.').lower()
    if fmt not in supported_formats['mol']:
        raise RuntimeError(f"The input molecule file format \"{fmt}\" is not supported by molstar.")
    if fmt == 'cif': fmt = 'mmcif'
    if fmt == 'cifcore': fmt = 'cifCore'
    # processing preset
    if 'target' in preset.keys():
        if not isinstance(preset['target'], list): preset['target'] = [preset['target']]
        preset['target'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['target']]
    if 'focus' in preset.keys():
        if not isinstance(preset['focus'], list): preset['focus'] = [preset['focus']]
        preset['focus'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['focus']]
    if 'targets' in preset.keys():
        if not isinstance(preset['targets'], list): preset['targets'] = [preset['targets']]
        preset['targets'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['targets']]
    if 'glycosylation' in preset.keys():
        if not isinstance(preset['glycosylation'], list): preset['glycosylation'] = [preset['glycosylation']]
        preset['glycosylation'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['glycosylation']]
    if 'colors' in preset.keys():
        if not isinstance(preset['colors'], list): preset['colors'] = [preset['colors']]
        for color in preset['colors']:
            if not isinstance(color['targets'], list): color['targets'] = [color['targets']]
            color['targets'] = [t.to_dict() if isinstance(t, Target) else t for t in color['targets']]
    d = {
        "type": 'mol',
        "data": data,
        "format": fmt,
        "preset": preset
    }
    if component: d['component'] = component
    return d

def parse_url(url, fmt=None, component=None, mol=True, preset={'kind': 'standard'}):
    """
    Parse the URL for `data` parameter of molstar viewer. 
    The url can be either a structure or a molstar state/session file. If a state/session
    was provided, the `mol` parameter should be set to `False`.

    Parameters
    ----------
    `url` — str
        The URL to the content to be passed to molstar.
    `fmt` — str (optional)
        Format of the input content. (default: `None`)

        Supported formats for structures include `cif`, `cifcore`, `pdb`, `pdbqt`, `gro`, `xyz`, `mol`, `sdf`, `mol2`, `lammps_data`, `lammps_traj_data`

        Supported formats for states and sessions include `json`, `molj`, `molx`, `zip`

        Supported formats for coordinates include `dcd`, `xtc`, `trr`, `nctraj`, `lammpstrj`
    `component` — dict | List[dict] (optional)
        Component to be created in molstar. 
        If not specified, molstar will use its default settings. (default: `None`)

        Use helper function `create_component` to generate correct data for this parameter.
    `mol` — DEPRECATED
    `preset` — dict (optional)
        The preset for molstar of how to display the loaded structure file.

    Returns
    -------
    `dict`
        The value for the `data` parameter.
    """
    # try to automatically infer file format
    parsed_url = urlparse(url)
    if not fmt:
        name, fmt = os.path.splitext(parsed_url.path)
    if not fmt:
        raise RuntimeError("The format must be specified if you are not providing static resources.")
    # check whether the format is in supported list and specify the url type
    fmt = fmt.strip('.').lower()
    for type, formats in supported_formats.items():
        if fmt in formats:
            urlfor = type
            break
    else:
        raise RuntimeError(f"The input file format \"{fmt}\" is not supported by molstar.")
    if fmt == 'cif': fmt = 'mmcif'
    if fmt == 'cifcore': fmt = 'cifCore'
    # processing preset
    if 'target' in preset.keys():
        if not isinstance(preset['target'], list): preset['target'] = [preset['target']]
        preset['target'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['target']]
    if 'focus' in preset.keys():
        if not isinstance(preset['focus'], list): preset['focus'] = [preset['focus']]
        preset['focus'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['focus']]
    if 'targets' in preset.keys():
        if not isinstance(preset['targets'], list): preset['targets'] = [preset['targets']]
        preset['targets'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['targets']]
    if 'glycosylation' in preset.keys():
        if not isinstance(preset['glycosylation'], list): preset['glycosylation'] = [preset['glycosylation']]
        preset['glycosylation'] = [t.to_dict() if isinstance(t, Target) else t for t in preset['glycosylation']]
    if 'colors' in preset.keys():
        if not isinstance(preset['colors'], list): preset['colors'] = [preset['colors']]
        for color in preset['colors']:
            if not isinstance(color['targets'], list): color['targets'] = [color['targets']]
            color['targets'] = [t.to_dict() if isinstance(t, Target) else t for t in color['targets']]
    d = {
        "type": 'url',
        "urlfor": urlfor,
        "data": url,
        "format": fmt,
        "preset": preset
    }
    if component: d['component'] = component
    return d

def parse_coordinate(inp, fmt=None):
    """
    Parse the coordinate file for loading a structure. This method encode the binary coordinate file
    into string with base64, so it is not recommended if you are about load a trajectory that is larger
    than 10 MB. For loading biger trajectories, try passing a url to molstar.

    Parameters
    ----------
    `inp` — str | file-like object
        The file path to molecule or the file content of molecule.

        If `inp` was set to file path, the format can be automatically determined.
        Otherwise the format has to be specified manually.
    `fmt` — str (optional)
        Format of the input molecule. 
        Supported formats include `dcd`, `xtc`, `trr`, `nctraj`, `lammpstrj` (default: `None`)

    Returns
    -------
    `dict`
        The value for the `coordinate` parameter of helper function `get_trajectory()`.

    Raises
    ------
    `RuntimeError`
        If the input format is not supported by molstar viewer, raises RuntimeError.
    """
    # provided a filename as input
    if os.path.isfile(inp):
        # if format is not specified, infer from filename
        if not fmt:
            name, fmt = os.path.splitext(inp)
        with open(inp, 'rb') as f:
            data = f.read()
    else:
        # provided a file-like object as input
        if isinstance(inp, IOBase):
            data = inp.read()
        # provided file content as input
        else:
            data = inp
    assert type(data) == bytes
    if not fmt: raise RuntimeError("The format must be specified if you didn't provide a file name.")
    fmt = fmt.strip('.').lower()
    if fmt not in supported_formats['coords']:
        raise RuntimeError(f"The input coordinate file format \"{fmt}\" is not supported by molstar.")
    return {
        'type': 'coord',
        "format": fmt,
        "data": base64.b64encode(data).decode('utf-8')
    }

def get_trajectory(topology, coordinate):
    """
    Load a trajectory into molstar viewer.

    Parameters
    ----------
    `topology` — dict
        The topology of molecule. Generated with helper function `parse_molecule()` or `parse_url()`
    `coordinate` — dict
        The coordinates of the trajectory. Generated with helper function `parse_coordinate()` or `parse_url()`

    Returns
    -------
    `dict`
        The value for the `data` parameter.
    """
    return {
        'type': 'traj',
        'topo': topology,
        'coords': coordinate
    }

def get_box(min_xyz=(0,0,0), max_xyz=(1,1,1), radius=0.1, label="Bounding Box", color='red', opacity=1.0):
    """
    Generate a bounding box in the viewer with given parameters.

    Parameters
    ----------
    `min_xyz` — tuple (optional)
        Minimum of x, y and z values (default: `(0,0,0)`)
    `max_xyz` — tuple (optional)
        Maximum of x, y and z values (default: `(1,1,1)`)
    `radius` — float (optional)
        Edge radius in angstrom (default: `0.1`)
    `label` — str (optional)
        The box label to be shown in the viewer (default: `"Bounding Box"`)
    `color` — str (optional)
        X11 color names (default: `'red'`)
        Avaliable options can be found at [here](https://www.w3.org/TR/css-color-3/#svg-color)
    `opacity` — float (optional)
        Transparency of the box. The value is ranging from 0 to 1.0. (default: `1.0`)

    Returns
    -------
    `dict`
        Dict for the `data` parameter of MolstarViewer

    Raises
    ------
    `ValueError`
        Raised if input coordinates are not 3-dimensional
    """
    if len(min_xyz) != 3: raise ValueError("Coordinates must be 3-dimensional!")
    if len(max_xyz) != 3: raise ValueError("Coordinates must be 3-dimensional!")
    return {
        'type': 'shape',
        'shape': 'box',
        'min': min_xyz,
        'max': max_xyz,
        'radius': radius,
        'label': label,
        'color': color,
        'alpha': opacity
    }

def get_sphere(center=(0,0,0), radius=1.0, label="Sphere", color='blue', opacity=1.0, detail=6):
    """
    Generate a sphere in the viewer with given parameters.

    Parameters
    ----------
    `center` — tuple (optional)
        Center of the sphere (default: `(0,0,0)`)
    `radius` — float (optional)
        Sphere radius in angstrom (default: `0.1`)
    `label` — str (optional)
        The sphere label to be shown in the viewer (default: `"Sphere"`)
    `color` — str (optional)
        X11 color names (default: `'blue'`)
        Avaliable options can be found at [here](https://www.w3.org/TR/css-color-3/#svg-color)
    `opacity` — int (optional)
        Transparency of the box. The value is ranging from 0 to 1.0. (default: `1.0`)
    `detail` — int (optional)
        Controls the subdivision surface of the sphere. The sphere is make of polygons. The higher
        the value, the more it looks like a fine sphere. But also requires longer time to
        render. The recommended value is 6.

    Returns
    -------
    `dict`
        Dict for the `data` parameter of MolstarViewer

    Raises
    ------
    `ValueError`
        Raised if input coordinates are not 3-dimensional
    """
    if len(center) != 3: raise ValueError("Coordinates must be 3-dimensional!")
    return {
        'type': 'shape',
        'shape': 'sphere',
        'center': center,
        'radius': radius,
        'label': label,
        'color': color,
        'alpha': opacity,
        'detail': detail
    }

def get_targets(chain, residue=None, atom=None, auth=False):
    """
    Select residues from a given chain. If no residue was specified, the entire chain will be selected.

    Parameters
    ----------
    `chain` — str
        Name of the target chain
    `residue` — int | List[int] (optional)
        Residue index of the target residues, corresponding to the structure file. (default: `None`)
    `atom` — int | List[int] (optional)
        Index of the target atom(s), corresponding to the structure file but started from 0. (default: `None`)
    `auth` — bool (optional)
        If a cif file was loaded, set `auth` to `True` to select the authentic chain names and residue numbers (default: `False`)

    Returns
    -------
    `Target`
        Selected chains, residues or atoms.
    """
    target = Target()
    target.auth = auth
    target.add_chain(chain)
    if residue is not None:
        if type(residue) != list: residue = [residue]
        for res in residue:
            if type(res) == int: target.chains[-1].add_residue(res)
            elif type(res) == str:
                try:
                    num = eval(res)
                    target.chains[-1].add_residue(num)
                except:
                    num = eval(res[:-1])
                    ins = res[-1]
                    target.chains[-1].add_residue(num, ins_code=ins)
            if atom is not None:
                if type(atom) != list: atom = [atom]
                for a in atom:
                    target.chains[-1].residues[-1].add_atom(a)
    return target

def create_component(label, targets, representation=Representation()):
    """
    Generate the component information for selected targets.

    Parameters
    ----------
    `label` — str
        Name of the component
    `targets` — Target | List[Target]
        List of targets, whose value should be generated by helper function `get_targets`
    `representation` — Representation | List[Representation] (optional)
        The representation(s) for this component (default: cartoon)

    Returns
    -------
    `dict`
        Component information that can be passed to function `parse_molecule`

    Raises
    ------
    `RuntimeError`
        If specified an invalid representation, raises `RuntimeError`
    """
    if type(targets) != list: targets = [targets]
    if type(representation) != list: representation = [representation]
    return {
        'label': label,
        'targets': [t.to_dict() if isinstance(t, Target) else t for t in targets],
        'representation': [r.to_dict() if isinstance(r, Representation) else r for r in representation]
    }

def get_selection(targets, select=True, add=False):
    """
    Select specific targets in the molstar viewer. The returned value can be passed to either `selection` or 
    `hover` parameters.

    Parameters
    ----------
    `targets` — Target | List[Target]
        List of targets, whose value should be generated by helper function `get_targets`

    `select`
        DEPRECATED

    `add` — bool (optional)
        If set to False, the viewer will clear the selections in corresponding mode before adding new selections.
        Otherwise the new selections will be added to existed ones. (default: `False`)

    Returns
    -------
    `dict`
        Selection data for callbacks.
    """
    if add: modifier = 'add'
    else: modifier = 'set'
    if type(targets) != list: targets = [targets]
    return {
        'targets': [t.to_dict() if isinstance(t, Target) else t for t in targets],
        'modifier': modifier
    }

def get_focus(targets, analyse=False):
    """
    Let the camera focus on the specified targets. 
    If `analyse` were set to true, non-covalent interactions within 5 angstroms will be analyzed.

    Parameters
    ----------
    `targets` — Target | List[Target]
        List of targets, whose value should be generated by helper function `get_targets`
    `analyse` — bool (optional)
        Whether to analyse the non-covalent interactions of targets to its surroundings within 5 angstroms  (default: `False`)

    Returns
    -------
    `dict`
        Focus data for callbacks
    """
    if type(targets) != list: targets = [targets]
    return {
        'targets': [t.to_dict() if isinstance(t, Target) else t for t in targets],
        'analyse': analyse
    }

def get_measurement(targets, type='label', options=None, add=False):
    """
    Create a measurement for the specified targets. There are 6 types of measurements supported by molstar, 
    specify them in the `type` parameter.

    Parameters
    ----------
    `targets` — Target | List[Target]
        The targets to be measured. The number of targets must meet the requirements of the specified measurement type. 
        The targets should be generated by helper function `get_targets`.
    `type` — str (optional)
        The 6 supported measurements are: label, orientation, plane, distance, angle, and dihedral. 
        For label, orientation, and plane, at least one target should be provided. For distances, angles, and dihedrals,
         2, 3, and 4 targets are needed respectively. (default: `'label'`)
    `add` — bool (optional)
        If set to False, existing measurements will be cleared before adding new ones.
        Otherwise the new measurements will be added to molecule. (default: `False`)

    Returns
    -------
    `dict`
        The measurement data for callbacks.

    Raises
    ------
    `ValueError`
        If the specified measurement type is not supported or the number of targets does not meet the requirements.
    `TypeError`
        If the targets are not valid `Target` objects, raises `TypeError`.
    """
    if not isinstance(targets, list): targets = [targets]
    minimum_required = {
        'label': 1,
        'orientation': 1,
        'plane': 1,
        'distance': 2,
        'angle': 3,
        'dihedral': 4
    }
    if type not in minimum_required.keys():
        raise ValueError(f"Invalid measurement type \"{type}\". Supported types are {minimum_required.keys()}.")
    if len(targets) < minimum_required[type]:
        raise ValueError(f"At least {minimum_required[type]} required by the measurement \"{type}\", only {len(targets)} provided.")

    for t in range(minimum_required[type]):
        if not isinstance(targets[t], Target) or not targets[t].valid:
            raise TypeError(f"Target {t} is not a valid Target object. Use helper function `get_targets` to generate targets.")
    return {
        'targets': [t.to_dict() if isinstance(t, Target) else t for t in targets],
        'type': type,
        'mode': 'add' if add else 'set',
    }


# AUTO GENERATED FILE - DO NOT EDIT

import typing  # noqa: F401
from typing_extensions import TypedDict, NotRequired, Literal # noqa: F401
from dash.development.base_component import Component, _explicitize_args

ComponentType = typing.Union[
    str,
    int,
    float,
    Component,
    None,
    typing.Sequence[typing.Union[str, int, float, Component, None]],
]

NumberType = typing.Union[
    typing.SupportsFloat, typing.SupportsInt, typing.SupportsComplex
]


class MolstarViewer(Component):
    """A MolstarViewer component.
The Molstar viewer component for dash

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- className (string; optional):
    The HTML property `class` for additional class names of the
    container of molstar viewer.

- data (boolean | number | string | dict | list; optional):
    Data containing the structure info that should be loaded into
    molstar viewer, as well as some control flags. The data can be
    generated with python method `parse_molecule`.

- focus (dict; optional):
    The structure region to let the camera focus on in the molstar
    viewer.

- frame (number; optional):
    The trajectory frame in the molstar viewer.

- hover (dict; optional):
    The structure region to be hovered in the molstar viewer.

- layout (dict; optional):
    The layout of the molstar viewer. Determining what controls to be
    displayed.   The layout is not allowed to be changed once the
    component has been initialized.

- measurement (boolean | number | string | dict | list; optional):
    The measurements in the molstar viewer.

- selection (dict; optional):
    The structure region to be selected in the molstar viewer.

- updatefocusonframechange (boolean; optional):
    Update focus data when frame index have changed.

- updateselectiononframechange (boolean; optional):
    Update selection data when frame index have changed."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_molstar'
    _type = 'MolstarViewer'


    def __init__(
        self,
        id: typing.Optional[typing.Union[str, dict]] = None,
        style: typing.Optional[typing.Any] = None,
        className: typing.Optional[str] = None,
        data: typing.Optional[typing.Any] = None,
        layout: typing.Optional[dict] = None,
        selection: typing.Optional[dict] = None,
        hover: typing.Optional[dict] = None,
        focus: typing.Optional[dict] = None,
        frame: typing.Optional[NumberType] = None,
        measurement: typing.Optional[typing.Any] = None,
        updatefocusonframechange: typing.Optional[bool] = None,
        updateselectiononframechange: typing.Optional[bool] = None,
        **kwargs
    ):
        self._prop_names = ['id', 'className', 'data', 'focus', 'frame', 'hover', 'layout', 'measurement', 'selection', 'style', 'updatefocusonframechange', 'updateselectiononframechange']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'className', 'data', 'focus', 'frame', 'hover', 'layout', 'measurement', 'selection', 'style', 'updatefocusonframechange', 'updateselectiononframechange']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(MolstarViewer, self).__init__(**args)

setattr(MolstarViewer, "__init__", _explicitize_args(MolstarViewer.__init__))



import json
from typing import Optional, Dict, Any
from importlib import resources

class Representation(object):
    def __init__(self, type: str = 'cartoon', color: Optional[str] = None, size: Optional[str] = None):
        self._params = self._load_valid_params('representation_params.json')
        
        self._type = 'cartoon'
        self._color = None
        self._size = None
        self._type_params: Dict[str, Any] = {}
        self._color_params: Dict[str, Any] = {}
        self._size_params: Dict[str, Any] = {}
        
        self.type = type
        if color:
            self.color = color
        if size:
            self.size = size

    def _load_valid_params(self, filename: str) -> Dict[str, Any]:
        try:
            with resources.open_text('dash_molstar.utils', filename) as f:
                return json.load(f)
        except (ImportError, FileNotFoundError):
            raise FileNotFoundError(f"Unable to locate the parameter definition file.")
        except json.JSONDecodeError:
            raise json.JSONDecodeError(f"The parameter definition file has been broken, please reinstall the package.")

    @property
    def type(self) -> str:
        return self._type

    @type.setter
    def type(self, value: str):
        if value in self._params.get('type', []):
            self._type = value
            self._type_params = {}  # reset type_params
        else:
            raise ValueError(f"Invalid type: {value}. Must be one of {self._params.get('type', [])}")

    @property
    def color(self) -> Optional[str]:
        return self._color

    @color.setter
    def color(self, value: Optional[str]):
        if value is None or value in self._params.get('color', []):
            self._color = value
            self._color_params = {}  # reset color_params
        else:
            raise ValueError(f"Invalid color: {value}. Must be one of {self._params.get('color', [])} or None")

    @property
    def size(self) -> Optional[str]:
        return self._size

    @size.setter
    def size(self, value: Optional[str]):
        if value is None or value in self._params.get('size', []):
            self._size = value
            self._size_params = {}  # reset size_params
        else:
            raise ValueError(f"Invalid size: {value}. Must be one of {self._params.get('size', [])} or None")

    def set_type_params(self, params: Dict[str, Any]):
        valid_params = self._params['typeParams'][self._type]
        self._set_params(self._type_params, params, valid_params, 'type')

    def set_color_params(self, params: Dict[str, Any]):
        if not self._color:
            raise ValueError("Color must be set before setting color params")
        valid_params = self._params['colorParams'][self._color]
        self._set_params(self._color_params, params, valid_params, 'color')

    def set_size_params(self, params: Dict[str, Any]):
        if not self._size:
            raise ValueError("Size must be set before setting size params")
        valid_params = self._params['sizeParams'][self._size]
        self._set_params(self._size_params, params, valid_params, 'size')

    def _set_params(self, target: Dict[str, Any], params: Dict[str, Any], valid_params: list, param_type: str):
        for key, value in params.items():
            if key in valid_params:
                target[key] = value
            else:
                print(f"Warning: '{key}' is not a valid parameter for {param_type}. Ignoring this parameter.")

    @staticmethod
    def np(name, params=None):
        """
        Generate `NamedParams` for parameters

        Parameters
        ----------
        `name` — str
            Parameter name
        `params` — any
            Parameters

        Returns
        -------
        `dict`
            NamedParams
        """
        return {'name': name, 'params': params}
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dict for creating representation

        Returns
        -------
        `Dict[str, Any]`
            Representation dict
        """
        result = {'type': self._type}
        if self._type_params:
            result['typeParams'] = self._type_params
        if self._color:
            result['color'] = self._color
            if self._color_params:
                result['colorParams'] = self._color_params
        if self._size:
            result['size'] = self._size
            if self._size_params:
                result['sizeParams'] = self._size_params
        return result
    
    def save_config(self, filename: str):
        """
        Save current representation configuration into JSON file.

        Parameters
        ----------
        `filename` — str
            JSON filename to be saved
        """
        if not filename.endswith('.json'): filename+='.json' 
        config = self.to_dict()
        with open(filename, 'w') as f:
            json.dump(config, f, indent=2)

    @classmethod
    def from_config(cls, filename: str) -> 'Representation':
        """
        Load a JSON configuration file and construct a new representation instance

        Parameters
        ----------
        `filename` — str
            JSON filename to be loaded

        Returns
        -------
        `Representation`
            The representation instance
        """
        with open(filename, 'r') as f:
            config = json.load(f)
        
        instance = cls(
            type=config.get('type', 'cartoon'),
            color=config.get('color'),
            size=config.get('size')
        )
        
        if 'typeParams' in config:
            instance.set_type_params(config['typeParams'])
        if 'colorParams' in config:
            instance.set_color_params(config['colorParams'])
        if 'sizeParams' in config:
            instance.set_size_params(config['sizeParams'])
        
        return instance
    
if __name__ == '__main__':
    rep = Representation('cartoon')


from typing import Dict, List, Optional, Union
import numpy as np

class Box(object):
    def __init__(self, min: tuple[float] = None, max: tuple[float] = None):
        self.__min = min
        self.__max = max
        self.__center = None
        self.__size = None
        if self.__min is not None and self.__max is not None:
            self.__center = ((self.__min[0] + self.__max[0]) / 2, (self.__min[1] + self.__max[1]) / 2, (self.__min[2] + self.__max[2]) / 2)
            self.__size = ((self.__max[0] - self.__min[0]), (self.__max[1] - self.__min[1]), (self.__max[2] - self.__min[2]))

    @property
    def min(self):
        if self.__min is not None:
            return self.__min
        return None

    @property
    def max(self):
        if self.__max is not None:
            return self.__max
        return None

    @property
    def center(self):
        if self.__center is not None:
            return self.__center
        return None

    @property
    def size(self):
        if self.__size is not None:
            return self.__size
        return None

    @property
    def min_x(self):
        if self.__min is not None:
            return self.__min[0]
        return None

    @property
    def min_y(self):
        if self.__min is not None:
            return self.__min[0]
        return None

    @property
    def min_z(self):
        if self.__min is not None:
            return self.__min[0]
        return None

    @property
    def max_x(self):
        if self.__max is not None:
            return self.__max[0]
        return None

    @property
    def max_y(self):
        if self.__max is not None:
            return self.__max[1]
        return None

    @property
    def max_z(self):
        if self.__max is not None:
            return self.__max[2]
        return None

    @property
    def size_x(self):
        if self.__size is not None:
            return self.__size[0]
        return None

    @property
    def size_y(self):
        if self.__size is not None:
            return self.__size[1]
        return None

    @property
    def size_z(self):
        if self.__size is not None:
            return self.__size[2]
        return None

    @property
    def center_x(self):
        if self.__center is not None:
            return self.__center[0]
        return None

    @property
    def center_y(self):
        if self.__center is not None:
            return self.__center[1]
        return None

    @property
    def center_z(self):
        if self.__center is not None:
            return self.__center[2]
        return None

    def to_sphere(self) -> 'Sphere':
        if self.__center is not None and self.__size is not None:
            center = self.__center
            radius = np.max(self.__size) / 2
            return Sphere(center, radius)
        return None

class Sphere(object):
    def __init__(self, center: tuple[float] = None, radius: float = None):
        self.__center = center
        self.__radius = radius

    @property
    def center(self):
        if self.__center is not None:
            return self.__center
        return None

    @property
    def radius(self):
        if self.__radius is not None:
            return self.__radius
        return None

    @property
    def center_x(self):
        if self.__center is not None:
            return self.__center[0]
        return None

    @property
    def center_y(self):
        if self.__center is not None:
            return self.__center[1]
        return None

    @property
    def center_z(self):
        if self.__center is not None:
            return self.__center[2]
        return None

    @property
    def radius_x(self):
        if self.__radius is not None:
            return self.__radius
        return None

    @property
    def radius_y(self):
        if self.__radius is not None:
            return self.__radius
        return None

    @property
    def radius_z(self):
        if self.__radius is not None:
            return self.__radius
        return None

    def to_box(self) -> Box:
        if self.__center is not None and self.__radius is not None:
            min = (self.__center[0] - self.__radius, self.__center[1] - self.__radius, self.__center[2] - self.__radius)
            max = (self.__center[0] + self.__radius, self.__center[1] + self.__radius, self.__center[2] + self.__radius)
            return Box(min, max)
        return None

class Boundary(object):
    def __init__(self, coords: np.ndarray = None):
        self.__box = None
        self.__sphere = None
        try:
            if coords is not None:
                min = (np.min(coords[:, 0]), np.min(coords[:, 1]), np.min(coords[:, 2]))
                max = (np.max(coords[:, 0]), np.max(coords[:, 1]), np.max(coords[:, 2]))
                center = ((min[0] + max[0]) / 2, (min[1] + max[1]) / 2, (min[2] + max[2]) / 2)
                radius = np.max(np.linalg.norm(coords - center, axis=1))
                self.__box = Box(min, max)
                self.__sphere = Sphere(center, radius)
        except:
            raise ValueError("Invalid coordinates provided to Boundary object")

    @property
    def box(self):
        return self.__box

    @property
    def sphere(self):
        return self.__sphere


class Atom(object):
    def __init__(self, index: int = None, name: str = None, x: float = None, y: float = None, z: float = None):
        self.__name = name
        self.__index = index
        self.__x = round(x, 6) if x is not None else None
        self.__y = round(y, 6) if y is not None else None
        self.__z = round(z, 6) if z is not None else None
        self.__valid = self.__index is not None

    @property
    def valid(self):
        return self.__valid

    @property
    def name(self):
        if not self.valid: return None
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    @property
    def index(self):
        if not self.valid: return None
        return self.__index

    @index.setter
    def index(self, value):
        if isinstance(value, str):
            try:
                int_value = int(value)
                self.__index = int_value
                self.__valid = True
            except Exception:
                raise ValueError(f"Invalid number value: {value}")
        elif isinstance(value, int):
            self.__index = value
            self.__valid = True
        else:
            raise TypeError("number must be a string or integer")

    @property
    def x(self):
        return self.__x

    @property
    def y(self):
        return self.__y

    @property
    def z(self):
        return self.__z

class Residue(object):
    def __init__(self, index: int = None, number: int = None, ins_code: str = '', name: str = '', atoms: List['Atom'] = []):
        self.__name = name
        self.__index = index
        self.__number = number
        if index is not None and number is None:
            self.__number = index
        self.__ins_code = ins_code
        self.__atoms = []
        self.__parse_data(atoms)
        self.__valid = (self.__number is not None) or (self.__index is not None)

    def __len__(self):
        return len(self.__atoms)

    def __parse_data(self, atoms: List['Atom']):
        if atoms:
            for atom in atoms:
                name = atom.get('name', '')
                index = atom.get('index')
                x = atom.get('x')
                y = atom.get('y')
                z = atom.get('z')
                self.add_atom(index, name, x, y, z)

    @property
    def valid(self):
        return self.__valid

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    @property
    def index(self):
        if not self.valid: return None
        return self.__index

    @index.setter
    def index(self, value):
        if isinstance(value, str):
            try:
                int_value = int(value)
                self.__index = int_value
                self.__valid = True
            except Exception:
                raise ValueError(f"Invalid number value: {value}")
        elif isinstance(value, int):
            self.__index = value
            self.__valid = True
        else:
            raise TypeError("number must be a string or integer")

    @property
    def number(self):
        if not self.valid: return None
        return self.__number

    @number.setter
    def number(self, value):
        if isinstance(value, str):
            try:
                int_value = int(value)
                self.__number = int_value
                self.__valid = True
            except Exception:
                raise ValueError(f"Invalid number value: {value}")
        elif isinstance(value, int):
            self.__number = value
            self.__valid = True
        else:
            raise TypeError("number must be a string or integer")

    @property
    def ins_code(self):
        return self.__ins_code

    @ins_code.setter
    def ins_code(self, value):
        self.__ins_code = value
        self.__valid = bool(value)

    @property
    def atoms(self) -> List[Atom]:
        return self.__atoms

    def find_atom(self, name: str) -> 'Atom':
        if not self.valid:
            raise ValueError('Cannot call find_atom on invalid Residue object')
        for atom in self.__atoms:
            if atom.name == name:
                return atom
        return Atom()

    def add_atom(self, index: int, name: str = None, x: float = None, y: float = None, z: float = None):
        atom = Atom(index, name, x, y, z)
        self.__atoms.append(atom)
    
    def remove_atom(self, name: Union[str, 'Atom']) -> bool:
        if isinstance(name, Atom):
            try:
                atom_index = self.__atoms.index(name)
                del self.__atoms[atom_index]
                return True
            except ValueError: pass
        else:
            atom = self.find_atom(name)
            if atom.valid:
                atom_index = self.__atoms.index(atom)
                del self.__atoms[atom_index]
                return True
        return False

class Chain(object):
    def __init__(self, chain_name: str = None, residues: List['Residue'] = [], auth_name: str = ''):
        self.__name = chain_name
        self.__auth_name = auth_name
        if chain_name and not auth_name:
            self.__auth_name = chain_name
        self.__residues = []
        self.__parse_data(residues)
        self.__valid = (self.__name is not None) or (self.__auth_name is not None)

    def __len__(self):
        return len(self.__residues)

    def __parse_data(self, residues: List['Residue']):
        if residues:
            for residue in residues:
                name = residue.get('name', '')
                index = residue.get('index')
                number = residue.get('number')
                ins_code = residue.get('ins_code', '')
                atoms = residue.get('atoms', [])
                self.add_residue(index, number, ins_code, name, atoms)

    @property
    def valid(self):
        return self.__valid

    @property
    def name(self):
        if not self.valid: return None
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value
        self.__valid = bool(value)

    @property
    def auth_name(self):
        if not self.valid: return None
        return self.__auth_name

    @auth_name.setter
    def auth_name(self, value):
        self.__auth_name = value
        self.__valid = bool(value)

    @property
    def residues(self) -> List[Residue]:
        return self.__residues

    @property
    def atoms(self) -> List[Atom]:
        atoms = []
        for residue in self.__residues:
            for atom in residue.atoms:
                if atom.valid:
                    atoms.append(atom)
        return atoms

    def find_residue(self, number, ins_code="") -> 'Residue':
        if not self.valid:
            raise ValueError('Cannot call find_residue on invalid Chain object')
        for residue in self.__residues:
            if residue.number == number and (residue.ins_code or "") == ins_code:
                return residue
        return Residue()

    def find_atom(self, residue_number, atom_name, ins_code="") -> 'Atom':
        if not self.valid:
            raise ValueError('Cannot call find_atom on invalid Chain object')
        residue = self.find_residue(residue_number, ins_code)
        if residue and residue.valid:
            return residue.find_atom(atom_name)
        return Atom()

    def add_residue(self, index: int, number: int = None, ins_code: str = '', name: str = '', atoms: List['Atom'] = []):
        residue = Residue(index, number, ins_code, name, atoms)
        self.__residues.append(residue)
    
    def remove_residue(self, number: Union[int, 'Residue'], ins_code: str = '') -> bool:
        if isinstance(number, Residue):
            try:
                residue_index = self.__residues.index(number)
                del self.__residues[residue_index]
                return True
            except ValueError: pass
        else:
            residue = self.find_residue(number, ins_code)
            if residue.valid:
                residue_index = self.__residues.index(residue)
                del self.__residues[residue_index]
                return True
        return False

class Target(object):
    def __init__(self, data: Dict = {}):
        self.__chains = []
        self.auth = False
        self.__parse_data(data)
        self.__boundary = None
        self.__valid = len(self.__chains) > 0

    def __len__(self):
        return len(self.__chains)

    def __parse_data(self, data: Dict):
        chains = data.get('chains', [])
        if chains:
            for chain in chains:
                chain_name = chain.get('name')
                auth_name = chain.get('auth_name', '')
                residues = chain.get('residues', [])
                self.add_chain(chain_name, residues, auth_name)

    @property
    def valid(self):
        return self.__valid

    @property
    def chains(self) -> List[Chain]:
        return self.__chains

    @property
    def residues(self) -> List[Residue]:
        residues = []
        for chain in self.__chains:
            for residue in chain.residues:
                if residue.valid:
                    residues.append(residue)
        return residues

    @property
    def atoms(self) -> List[Atom]:
        atoms = []
        for chain in self.__chains:
            for residue in chain.residues:
                for atom in residue.atoms:
                    if atom.valid:
                        atoms.append(atom)
        return atoms

    @property
    def boundary(self) -> Optional[Boundary]:
        if self.valid:
            if not self.__boundary:
                coords = np.array([[atom.x, atom.y, atom.z] for atom in self.atoms])
                self.__boundary = Boundary(coords)
            return self.__boundary
        raise ValueError('Cannot access boundary on invalid Target object')

    def find_chain(self, name) -> 'Chain':
        if not self.__valid:
            raise ValueError('Cannot call find_chain on invalid Target object')
        for chain in self.__chains:
            if chain.name == name:
                return chain
        return Chain()

    def find_residue(self, chain_name, residue_number, ins_code="") -> 'Residue':
        chain = self.find_chain(chain_name)
        if chain.valid:
            for residue in chain.residues:
                if residue.number == residue_number and (residue.ins_code or "") == ins_code:
                    return residue
        return Residue()

    def find_atom(self, chain_name, residue_number, atom_name, ins_code="") -> 'Atom':
        chain = self.find_chain(chain_name)
        if chain.valid:
            residue = chain.find_residue(residue_number, ins_code)
            if residue.valid:
                return residue.find_atom(atom_name)
        return Atom()

    def add_chain(self, chain_name: str, residues: List = [], auth_name: str = ''):
        chain = Chain(chain_name, residues, auth_name)
        self.__chains.append(chain)
        self.__valid = True

    def remove_chain(self, chain_name: Union[str, 'Chain']) -> bool:
        if isinstance(chain_name, Chain):
            try:
                chain_index = self.__chains.index(chain_name)
                del self.__chains[chain_index]
                if not self.__chains:
                    self.__valid = False
                return True
            except ValueError: pass
        else:
            chain = self.find_chain(chain_name)
            if chain.valid:
                chain_index = self.__chains.index(chain)
                del self.__chains[chain_index]
                if not self.__chains:
                    self.__valid = False
                return True
        return False
    
    def to_dict(self) -> Dict:
        data = {
            'chains': [],
            'auth': self.auth
        }
        for chain in self.__chains:
            chain_data = {
                'name': chain.name,
                'auth_name': chain.auth_name,
                'residues': []
            }
            for residue in chain.residues:
                residue_data = {
                    'name': residue.name,
                    'index': residue.index,
                    'number': residue.number,
                    'ins_code': residue.ins_code,
                    'atoms': []
                }
                for atom in residue.atoms:
                    atom_data = {
                        'name': atom.name,
                        'index': atom.index,
                        'x': atom.x,
                        'y': atom.y,
                        'z': atom.z
                    }
                    residue_data['atoms'].append(atom_data)
                chain_data['residues'].append(residue_data)
            data['chains'].append(chain_data)
        return data

if __name__ == "__main__":
    sample_data = {'chains': [{'name': 'L', 'auth_name': 'L', 'residues': [{'name': 'GLY', 'index': 62, 'number': 62, 'ins_code': '', 'atoms': [{'name': 'N', 'index': 4817, 'x': 3.759000062942505, 'y': -11.343000411987305, 'z': -5.960999965667725}, {'name': 'CA', 'index': 4818, 'x': 4.876999855041504, 'y': -12.213000297546387, 'z': -6.27400016784668}, {'name': 'C', 'index': 4819, 'x': 6.201000213623047, 'y': -11.520999908447266, 'z': -6.026000022888184}, {'name': 'O', 'index': 4820, 'x': 6.288000106811523, 'y': -10.29699993133545, 'z': -6.059000015258789}]}, {'name': 'ASN', 'index': 133, 'number': 133, 'ins_code': '', 'atoms': [{'name': 'N', 'index': 5366, 'x': -15.53499984741211, 'y': -41.314998626708984, 'z': 11.765999794006348}, {'name': 'CA', 'index': 5367, 'x': -16.469999313354492, 'y': -40.20199966430664, 'z': 11.64799976348877}, {'name': 'C', 'index': 5368, 'x': -16.136999130249023, 'y': -38.9630012512207, 'z': 12.482000350952148}, {'name': 'O', 'index': 5369, 'x': -15.72700023651123, 'y': -39.069000244140625, 'z': 13.640000343322754}, {'name': 'CB', 'index': 5370, 'x': -17.892000198364258, 'y': -40.66999816894531, 'z': 11.961000442504883}, {'name': 'CG', 'index': 5371, 'x': -18.92799949645996, 'y': -39.95800018310547, 'z': 11.12399959564209}, {'name': 'OD1', 'index': 5372, 'x': -18.735000610351562, 'y': -39.75400161743164, 'z': 9.925000190734863}, {'name': 'ND2', 'index': 5373, 'x': -20.030000686645508, 'y': -39.5620002746582, 'z': 11.75100040435791}]}]}, {'name': 'H', 'auth_name': 'H', 'residues': [{'name': 'ARG', 'index': 105, 'number': 99, 'ins_code': 'B', 'atoms': [{'name': 'N', 'index': 825, 'x': 1.0479999780654907, 'y': -5.711999893188477, 'z': -34.37200164794922}, {'name': 'CA', 'index': 826, 'x': 2.497999906539917, 'y': -5.5329999923706055, 'z': -34.34299850463867}, {'name': 'C', 'index': 827, 'x': 3.2239999771118164, 'y': -6.763000011444092, 'z': -33.80699920654297}, {'name': 'O', 'index': 828, 'x': 4.064000129699707, 'y': -7.335999965667725, 'z': -34.49399948120117}, {'name': 'CB', 'index': 829, 'x': 2.881999969482422, 'y': -4.296999931335449, 'z': -33.5260009765625}, {'name': 'CG', 'index': 830, 'x': 2.4560000896453857, 'y': -2.9779999256134033, 'z': -34.152000427246094}, {'name': 'CD', 'index': 831, 'x': 2.986999988555908, 'y': -1.7869999408721924, 'z': -33.358001708984375}, {'name': 'NE', 'index': 832, 'x': 4.429999828338623, 'y': -1.6050000190734863, 'z': -33.513999938964844}, {'name': 'CZ', 'index': 833, 'x': 4.988999843597412, 'y': -0.7549999952316284, 'z': -34.37200164794922}, {'name': 'NH1', 'index': 834, 'x': 4.22599983215332, 'y': -0.003000000026077032, 'z': -35.154998779296875}, {'name': 'NH2', 'index': 835, 'x': 6.309999942779541, 'y': -0.6520000100135803, 'z': -34.446998596191406}]}, {'name': 'GLY', 'index': 116, 'number': 108, 'ins_code': '', 'atoms': [{'name': 'N', 'index': 942, 'x': -10.392000198364258, 'y': -21.732999801635742, 'z': -22.767000198364258}, {'name': 'CA', 'index': 943, 'x': -11.680999755859375, 'y': -22.361000061035156, 'z': -22.992000579833984}, {'name': 'C', 'index': 944, 'x': -12.182999610900879, 'y': -22.957000732421875, 'z': -21.683000564575195}, {'name': 'O', 'index': 945, 'x': -11.437999725341797, 'y': -23.02899932861328, 'z': -20.70199966430664}]}, {'name': 'CYS', 'index': 204, 'number': 200, 'ins_code': '', 'atoms': [{'name': 'N', 'index': 1558, 'x': -31.833999633789062, 'y': -45.24300003051758, 'z': 0.6779999732971191}, {'name': 'CA', 'index': 1559, 'x': -30.711000442504883, 'y': -44.742000579833984, 'z': -0.10599999874830246}, {'name': 'C', 'index': 1560, 'x': -31.20800018310547, 'y': -44.34600067138672, 'z': -1.4950000047683716}, {'name': 'O', 'index': 1561, 'x': -31.722000122070312, 'y': -45.1870002746582, 'z': -2.2300000190734863}, {'name': 'CB', 'index': 1562, 'x': -29.65999984741211, 'y': -45.847999572753906, 'z': -0.23800000548362732}, {'name': 'SG', 'index': 1563, 'x': -27.940000534057617, 'y': -45.314998626708984, 'z': -0.17599999904632568}]}]}]}
    target = Target(sample_data)
    print("Target valid:", target.valid)
    print("Number of chains:", len(target))
    for chain in target.chains:
        print("Chain name:", chain.name, ",Auth name:", chain.auth_name)
        for residue in chain.residues:
            print("  Residue index:", residue.index, ",Residue number:", str(residue.number)+residue.ins_code, ",Name:", residue.name)
            for atom in residue.atoms:
                print("    Atom name:", atom.name, ",x:", atom.x, ",y:", atom.y, ",z:", atom.z)
    print("Number of chains:", len(target))
    print("Number of residues:", len(target.residues))
    print("Number of atoms:", len(target.atoms))
    print("Finding atom:", target.find_atom('H', 99, 'N', 'B').valid) # True
    print("Finding residue:", target.find_residue('L', 71).valid) # False
    print("Finding chain:", target.find_chain('L').valid)  # True
    print("Finding chain:", target.find_chain('X').valid)  # False
    print("Finding residue:", target.find_residue('X', 71).valid)  # False
    residue = target.find_residue('H', 99, 'B')
    CA = residue.find_atom('CA')
    CB = residue.find_atom('CB')
    print("Finding atom CA:", CA.valid) # True
    print("Finding atom CB:", CB.valid) # True
    residue.remove_atom('CA')
    residue.remove_atom(CB)
    print("Finding atom CA after remove:", residue.find_atom('CA').valid) # False
    print("Finding atom CB after remove:", target.find_atom('H', 99, 'CB', 'B').valid) # False
    print("Boundary box min:", target.boundary.box.min)
    print("Boundary box max:", target.boundary.box.max)
    print("Boundary sphere center:", target.boundary.sphere.center)
    print("Boundary sphere radius:", target.boundary.sphere.radius)